/* -----------------------------------------------------------------------------
*
*  npmodels.c
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2016 by Shane Saxon - saxon@openantz.com
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named LICENSE.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#include "npmodels.h"
//#include "npassimp.h"
#include "../gl/npgeometry.h" // lv, models
#include "../gl/nptexmap.h" // lv, models
#include "../npfile.h" // lv, models
#include "../../data/npmapfile.h"

#include "../../npdata.h"

void npTexMapSync2(void* dataRef)
{
	pData data = (pData) dataRef;
	pNPtexmap tmptexmap = NULL;
	pNPtexmap texmap = NULL;
	int i, x = 0;
	int z = 0;

	/*
		For each tmptexmap
			check every entry in texmap to see if tmptexmap is already in there under a different extId
				if yes, reassign tmptexmap->extId to the already used extId of texmap.
				if no, call npTex(kNPtexlistAddFile, &x, tmptexmap->filename, tmptexmap->path, dataRef);
				
	*/
	for(i = 1; i < 2000; i++)
	{
		tmptexmap = &data->io.gl.tmptexmap[i];
		if(tmptexmap->extTexId == 0)
			continue;

		for(x = 1; x < 2000; x++)
		{
			texmap = data->io.gl.texmap[x];
			if(texmap)
			{
				printf("MAP %s %s\n\n", texmap->filename, texmap->path);
				if( (strcmp(texmap->filename, tmptexmap->filename) == 0) &&
					(strcmp(texmap->path, tmptexmap->path)		   == 0) )
				{
					// Same Mapping in tmptexMap and texMap
					printf("Same\n");
				//	tmptexmap->extTexId = texmap->extTexId; printf("same\n");
					break;
				}
				else
				{
					// Different Mapping
					//printf("Diff\n");
					z = 1;
				}
			}
		}

		if(x == 2000 && tmptexmap->extTexId > 0)
		{
		//	printf("No Mapping\n");
			for(x = 1; x < 2000; x++)
			{
				texmap = data->io.gl.texmap[x];
				if(!texmap)
				{
					printf("5555 TMP %s %s\n\n", tmptexmap->filename, tmptexmap->path); //system("pause");
					texmap = npTex(kNPtexlistAddFile, &x, tmptexmap->filename, tmptexmap->path, dataRef);	
					tmptexmap->extTexId = x;
					break;
				}
			}
		}
	}
}

// npTexMapSync(void* dataRef); data->io.gl.tmpTexMap
void npTexMapSync(void* dataRef)
{
	pData data = (pData) dataRef;
	pNPtexmap tmptexmap = NULL;
	pNPtexmap texmap = NULL;
	int i = 0;
	int x = 0;

	for(i = 1; i < 2000; i++)
	{
		tmptexmap = &data->io.gl.tmptexmap[i];
		if(tmptexmap->extTexId == i)
		{
			texmap = data->io.gl.texmap[i];
			if(texmap)
			{
				if( (strcmp(texmap->filename, tmptexmap->filename) == 0) &&
					(strcmp(texmap->path, tmptexmap->path) == 0) )
				{
					// Same Mapping in tmptexMap and texMap
				//	break;
					continue;
				}
				else
				{	// Different Mapping in tmpGeoMap and GeoMap
					for(x = 1; x < 2000; x++)
					{
						texmap = data->io.gl.texmap[x];

						//npTexMapEmpty
						/*
						if(texmap && texmap->filename[0] == '\0' && texmap->path[0] ==  '\0')
						{
							texmap = npTex(kNPtexlistAddFile, &x, tmptexmap->filename, tmptexmap->path, dataRef);
							tmptexmap->extTexId = x;
							break;
						}
						*/
						if(!texmap)
						{
							texmap = npTex(kNPtexlistAddFile, &x, tmptexmap->filename, tmptexmap->path, dataRef);
							tmptexmap->extTexId = x;
							break;
						}

					}
				}

			}
			else
			{
				// No mapping
				printf("No mapping");
				texmap = npTex(kNPtexlistAddFile, &i, tmptexmap->filename, tmptexmap->path, dataRef);
				texmap->extTexId = tmptexmap->extTexId;
				//	geo = npGeo(kNPsetGeoId, &i, &tmpGeo->textureId, tmpGeo->modelPath, tmpGeo->modelFile, dataRef);
				//	geo->extTexId = tmpGeo->extTexId;
		}
		}

	}
}

// npGeoMapNew
/*
	Syncs/Merges the tmpGeoMap into the GeoMap

*/
void npGeoSync (void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeo tmpGeo = NULL;
	pNPgeo geo = NULL;
	int i;
	int x;
	int texId;

	for(i = 1001; i < 2000; i++)
	{
		tmpGeo = &data->io.gl.tmpGeoMap[i];
		if(tmpGeo->geometryId == i) // checks for assignment
		{
			geo = data->io.gl.geoMap[i];
			if(geo)
			{
				if( (strcmp(geo->modelFile, tmpGeo->modelFile) == 0) &&
					(strcmp(geo->modelPath, tmpGeo->modelPath) == 0) )
				{
					// Same Mapping in tmpGeoMap and GeoMap
				//	break;
					continue;
				}
				else
				{	// Different Mapping in tmpGeoMap and GeoMap
					for(x = 1001; x < 2000; x++)
					{
						geo = data->io.gl.geoMap[x];
						if(!geo)
						{
							printf("Found Unused Geo\n");
							printf("tmpGeo->modelFile : %s\n", tmpGeo->modelFile);
							printf("tmpGeo->modelPath : %s\n", tmpGeo->modelPath);
							//system("pause");
							geo = npGeo(kNPsetGeoId, &x, &tmpGeo->textureId, tmpGeo->modelPath, tmpGeo->modelFile, dataRef);
							npGeo(kNPgeoLoad, &x, &tmpGeo->textureId, tmpGeo->modelPath, tmpGeo->modelFile, dataRef);
							//geo = npGeo(kNPgeoLoad, &geo->geometryId, NULL, geo->modelPath, geo->modelFile, dataRef);
							tmpGeo->geometryId = x;
							break;
						}
					}
				}
			}
			else
			{
				// No mapping
				geo = npGeo(kNPsetGeoId, &i, &tmpGeo->textureId, tmpGeo->modelPath, tmpGeo->modelFile, dataRef);
				geo->extTexId = tmpGeo->extTexId;
			}
		}
	}
}

void npSyncGeos (void* dataRef)
{
	pData data = (pData) dataRef;

	int i = 0, j = 0;
	int count = 0;
	void** nodes;
	pNPnode node;

	data->map.syncNodes = false;		//reset sync flags
//	data->map.syncTagsReady = false;

						
	//allocate a pair of destination buffers for the nodes and tags
	nodes = npMalloc(0, sizeof(pNPnode) * kNPnodeMax, data);			//data->map.nodeCount
	if (!nodes) return;
//	tags = npMalloc(0, sizeof(pNPnode) * kNPnodeMax, data);			//update to use tag count //zz debug
//	if (!tags) return;

	//populate the sort lists, skip over null nodes
	for (i=0; i < kNPnodeMax; i++)	//data->map.nodeCount; i++)
		if (data->map.nodeID[i] != NULL)
			nodes[j++] = data->map.nodeID[i];

	//sort nodes and tags grouped first by table_id and second by record_id
	/*
	npSort(nodes, j, sizeof(void*), npCompareNodes, data);
	npSort(data->io.gl.hud.tags.list,
			data->io.gl.hud.tags.count,
			sizeof(void*), npCompareTags, data);	//update to use tag count //zz debug
	*/

	//loop through all nodes and update with tag
//	tagsIndex = 0;	//reset the index, takes advantage of sorted lists, zzhp
	for (i=0; i < j; i++)
	{	//npTagNode(nodes[i], data);
	node = (pNPnode)nodes[i];
	if(node->geometry > 1000)
	{
		printf("geo > 1000\n");
	//	data->io.gl.geoMap[node->geometry]->textureId
	}
	//printf("id: %d  tag: %s\n",  node->id, node->tag->title);
	}
	printf("tag count J: %d\n", j);
	//loop through all nodes and attach tagPtr based on recordID and tableID
//	for (i=0; i < data->map.nodeRootCount; i++)
//		npNodeTraverseTree (npTagNode, data->map.node[i], dataRef);

	npFree(nodes, data);
}


void npInitTmpGeoMap(void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geo = NULL;
	int i = 0;

	data->io.gl.loadGeos = 0;
	for(i = 0; i <= 2000; i++)
	{
		geo = &data->io.gl.tmpGeoMap[i];
		geo->center.x = 0; geo->center.y = 0; geo->center.z = 0;
		geo->extTexId = 0; geo->geometryId = 0; geo->index = 0; geo->loaded = 0;
		geo->modelFile[0] = '\0'; geo->modelFilePath[0] = '\0'; geo->modelId = 0;
		geo->modelPath[0] = '\0'; geo->modelTextureFile[0] = '\0'; geo->modelTextureFilePath[0] = '\0';
		geo->modelTexturePath[0] = '\0'; geo->name[0] = '\0';
		geo->modelTextureFilePath[0] = '\0';
		geo->rotate.x = 0; geo->rotate.y = 0; geo->rotate.z = 0;
		geo->scale.x = 0; geo->scale.y = 0; geo->scale.z = 0;
		geo->scene = NULL; geo->textureId = 0;
	}
}

void npInitGeoMap(void* dataRef);
void npInitGeoMap(void* dataRef)
{
	pData data = (pData) dataRef;
	int i = 0;
	

	for(i = 0; i <= 2000; i++)
	{
		data->io.gl.geoMap[i] = NULL;
	}
}

/** Inits the Geolist 
	@param dataRef is a global map reference instance.
*/
void npInitModels (void* dataRef)
{
	npInitGeoList(dataRef);
	npInitGeoMap(dataRef);
	npLoadGeos(dataRef);
	return;
}

void npLoadGeos(void* dataRef)
{
	int i = 0;
	int x = 0;
	int result = 0;
	int fileType = 0;
	int textureSize = 0;
	int fileNumber = 1;
	static int extTexId = 0;
	int geoId = 1001;
	int fileCat = 0;

	pNPfileRef fRef = NULL;

	char* filename = (char*)malloc(4096);
	char path[256] = {'\0'};
	char filePath[4096] = {'\0'}; // lv models

	//unsigned int textureID;		//zz debug, allow loading textures at runtime
									//detect changes to data->io.file.mapPath
	pData data = (pData) dataRef;
	NPfloatXYZ center, rotate, scale;
	
	center.x = 0; rotate.x = 0; scale.x = 0;
	center.y = 0; rotate.y = 0; scale.y = 0;
	center.z = 0; rotate.z = 0; scale.z = 0;


	printf ("Loading Geos... : %s\n", data->io.file.appPath);

	fRef = nposNewFileRef( data );

	result = nposFindFirstFile( fRef, "usr/global/models/", "*.*", data );
	if( result != 1 )
		return;		// err or empty folder

	do
    {
		i++;

		// print a few of the filenames then dots for every 100 files
		if( i <= 5 )
			printf( "%.70s\n", fRef->name );
		else if(  i < 100 || i % 100 == 0 )
			printf( "." );

		//printf("fRef->name : %s\n", fRef->name);
		sprintf(filename, "%s/%s", "usr/global/models/", fRef->name );

		// if Folder (not a file) then recursively call to create dir tree
		if( fRef->isDir )
		{
		//	npLoadTextures( sPath, data );			// recursion
		}
		else
		{
			sprintf(path, "%s%s", data->io.file.appPath, "usr\\global\\models\\");
		//	printf("---------123 path : %s--------\n", path);
			if( ( strcmp(fRef->name, "models-notes.txt") == 0) || (strcmp(fRef->name, ".DS_Store") == 0) || (strcmp(fRef->name, "README") == 0)  )
				continue;

			
			extTexId = 0;

			sprintf(filePath, "%s%s", path, fRef->name);
			npGetFileTypeCat( &fileCat, filePath, dataRef);
			if(fileCat != kNPfileCatImage)
			{
//				printf("npLoadGeos filePath : %s\n", filePath);	
				for(x = 0; x < strlen(path); x++)
					path[x] = tolower(path[x]);

				for(x = 0; x < strlen(fRef->name); x++)
					fRef->name[x] = tolower(fRef->name[x]);

				npGeo(kNPsetGeoId, &geoId, NULL, path, fRef->name, dataRef);
				geoId++;
			}	
		}
    }
	while( nposFindNextFile( fRef ) );	// next file within limits

    nposFindClose( fRef, data );		// always clean up!
}


/** Unloads all the models
	@param dataRef is a global map reference instance.	
*/
void npCloseModels (void* dataRef)
{
	/// @todo npCloseModels
	return;
}

// node->texture = npGeoTexId( node->geometry, dataRef );
int npGeoTexId( int geometryId, void* dataRef )
{
	pData data = (pData) dataRef;
	pNPgeolist geolist = &data->io.gl.geolist[0];
	int i = kNPgeoListMax;
	int geoId = geometryId;
	
	for(; i > 0; i--)
	{
		geolist = &data->io.gl.geolist[i];
		if(geolist->geometryId == geoId)
		{
			return geolist->textureId;
		}
	}

	return 0;
}

pNPgeolist npSearchGeolistExtTexId(int extTexId, void* dataRef);
pNPgeolist npSearchGeolistExtTexId(int extTexId, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geo = NULL;
	int i = 0;

	if(extTexId == 0)
		return geo;

	for(i = 1; i < kNPgeoListMax; i++)
	{
		geo = &data->io.gl.geolist[i];
		if(geo->textureId == extTexId)
		{
			return geo;
		}
	}

	geo = NULL;
	return geo; 
}

pNPgeolist npSearchGeolistModelFilePath(char* file, char* path, void* dataRef);
pNPgeolist npSearchGeolistModelFilePath(char* file, char* path, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geo = NULL;
	int i = 0;

	for(i = 0; i < kNPgeoListMax; i++)
	{
		geo = &data->io.gl.geolist[i];
		if( (strcmp(geo->modelFile, file) == 0) 
			&& (strcmp(geo->modelPath, path) == 0) )
		{
			return geo;
		}
	}

	return NULL;
}

pNPgeolist npGetUnusedGeo(void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geolist = NULL;
	pNPgeolist geodef = &data->io.gl.geolist[0];
	int i = 1;

	for(i = 1; i < kNPgeoListMax; i++)
	{
		geolist = &data->io.gl.geolist[i];
		if( 0 == memcmp(geodef, geolist, sizeof(NPgeolist)) )
			return geolist;
	}

	return NULL;
}

/// return a new GeoId
int npNewGeoId(void* dataRef);
int npNewGeoId(void* dataRef)
{
	static int geoId = 999;

	geoId++;
	printf("623982901 geo id %d\n", geoId);
	
	return geoId;
}

void npTexMapPrint(void* dataRef);
void npTexMapPrint(void* dataRef)
{
	pData data = (pData) dataRef;
	pNPtexmap texmap = data->io.gl.texmap[0];
	int i = 1;

	for(i = 1; i < 2000; i++)
	{
		texmap = data->io.gl.texmap[i];
		if(texmap->extTexId > 0)
			printf("\nextTexId : %d\nintTexId : %d\n", texmap->extTexId, texmap->intTexId);
	}
}


// 1/12/17 models start
int npModelNewId(pNPgeolist geo, void* dataRef);
int npModelNewId(pNPgeolist geo, void* dataRef)
{
	int modelId = geo->geometryId-1000;
	if( geo->geometryId <= 1000 )
		modelId = 0;

	return modelId;
}


const struct aiScene* npModelImportNew(pNPgeolist geo, void* dataRef);
const struct aiScene* npModelImportNew(pNPgeolist geo, void* dataRef)
{
	const struct aiScene* scene = NULL;
//	geo->
	scene = aiImportFile(geo->modelFilePath, 0);
//	scene = aiImportFile(geo->modelFilePath, aiProcess_GenNormals | aiProcess_FixInfacingNormals);

	if( scene == NULL )
		printf("\naiImportFile returned null");
	
	geo->scene = scene;

	return scene;
}

char* npModelTextureFile(pNPgeolist geo, void* dataRef);
char* npModelTextureFile(pNPgeolist geo, void* dataRef)
{
	pData data = (pData) dataRef;
//	((pNPassimp)data->io.assimp)->path
	const struct aiScene* scene = geo->scene;
	struct aiMaterial* material = NULL;
	char tempStr[1024] = {'\0'};
	int len = 0;
	int x = 0;
    static const int types[] = {
          aiTextureType_NONE,
          aiTextureType_DIFFUSE,
          aiTextureType_SPECULAR,
          aiTextureType_AMBIENT,
          aiTextureType_EMISSIVE,
          aiTextureType_HEIGHT,
          aiTextureType_NORMALS,
          aiTextureType_SHININESS,
          aiTextureType_OPACITY,
          aiTextureType_DISPLACEMENT,
          aiTextureType_LIGHTMAP,
          aiTextureType_REFLECTION,
          aiTextureType_UNKNOWN
    };
	unsigned int type = 0;
	unsigned int idx = 0;
	unsigned int total = 0;

//	geo->path.data[0] = '0';
	((pNPassimp)(data->io.assimp))->path.data[0] = '\0';

//	if(sceneIndex < 1 || sceneIndex > 
	/*
	if(sceneIndex < 1 || sceneIndex > assimp->numOfScenes)
	{
		printf("\nInvalid Scene Index");
		return assimp->path;
	}
	*/
	if( scene == NULL)
	{
		printf("\nScene not loaded");
		return NULL;
	}

	material = scene->mMaterials[0];
	
	for(type = 0; type < sizeof(types)/sizeof(types[0]); ++type)
	{
	//	printf("\ntype : %d", type);
		for(idx = 0; AI_SUCCESS==aiGetMaterialString(scene->mMaterials[0],
			AI_MATKEY_TEXTURE(types[type],idx),&((pNPassimp)(data->io.assimp))->path ); idx++)
		{
			printf("%s\n    \'%s\'",(total++?"":"\nTexture Refs:" ), ((pNPassimp)(data->io.assimp))->path.data);
			strcpy(tempStr, ((pNPassimp)(data->io.assimp))->path.data);
			total++;
		}
	}

	len = strlen(tempStr);
	while( --len > 0 )
	{
		if( (tempStr[len] == '\\') || (tempStr[len] == '/') )
		{
			strcpy(geo->modelTextureFile, &tempStr[len+1]);
			strncpy(geo->modelTexturePath, tempStr, len+1);
			break;
		}
	}
//	len = strlen( geo->modelTextureFile );
//	for(; x < len; x++)
	if(len == 0)
		strcpy(geo->modelTextureFile, tempStr);

	//printf("geo->modelTextureFile : %s\n", geo->modelTextureFile);
	//system("pause");
	return geo->modelTextureFile;
}

pNPtexmap npTexlistAddModelTex(pNPgeolist geo, void* dataRef);
pNPtexmap npTexlistAddModelTex(pNPgeolist geo, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPtexmap tex = NULL;
	//char* modelPath = geo->modelPath;
	
	// if the model's texture path is empty than first try the model's path
	if(geo->modelTexturePath[0] == '\0')
	{
		strcpy(geo->modelTexturePath, geo->modelPath);
		sprintf(geo->modelTextureFilePath, "%s%s", geo->modelTexturePath, geo->modelTextureFile);
	}

	tex = npTex(kNPtexlistAddFile, &geo->extTexId, geo->modelTextureFile, geo->modelTexturePath, dataRef);
	//tex = npTexlistAddFile(geo->modelTextureFile, geo->modelTexturePath, geo->extTexId, dataRef);
	return tex;
}

int npLoadModel(pNPgeolist geo, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPassimp assimp = (pNPassimp)data->io.assimp;
	pNPtexmap tex = NULL;
	int modelId = 0;
	char* texFileName = NULL;
	
	modelId = npModelNewId(geo, dataRef);
	assimp->scene[modelId] = npModelImportNew(geo, dataRef);
	//printf("File Path : %s\n", geo->modelFilePath);
	if(assimp->scene[modelId] == NULL)
	{
	//	printf("Failed importing 3d model\n"); /// @todo lv models, print to console
	//	printf("File Path : %s\n", geo->modelFilePath);
		return 0;
	}

	texFileName = npModelTextureFile(geo, dataRef);
	printf("Model Texture File : %s\n", texFileName);
//	tex = npTexlistSearchFile(texFileName, geo->modelTexturePath, dataRef);
	printf("Model Texture Path : %s\n", geo->modelTexturePath);
	tex = npTex(kNPtexlistSearchFile, NULL, texFileName, geo->modelTexturePath, dataRef);
	if(tex)
	{	
		if(tex->loaded == 1)
			geo->textureId = tex->intTexId;
		else
		{
		//	printf("TODO npTexlistAddFile From npLoadModel\n");
		//	npTexlistAddFile(texFileName, geo->modelTexturePath, dataRef);
		//	geo->textureId = tex->intTexId = npLoadTexture(, 0, dataRef);
		}
	}
	else
	{
		printf("File / Path Pair not already in the Texlist\n");
		if(texFileName[0] != '\0')
		{
			tex = npTexlistAddModelTex(geo, dataRef);
			geo->extTexId = tex->extTexId;
		}
	}

	return modelId;
}

// 1/12/17 models end

int npGeoNull(void)
{
	return 0;
}

char* npModelNewCenter(char* curs, float* x, float* y, float* z, void* dataRef);
char* npModelNewCenter(char* curs, float* x, float* y, float* z, void* dataRef)
{
	pNPgeolist geolist = NULL;

	(*x) = npstrtof(&curs);
	(*y) = npstrtof(&curs);
	(*z) = npstrtof(&curs);

	return curs;
}

char* npModelNewGeoId( char* idVal, int* geometryId, void* dataRef)
{
	pData data = (pData) dataRef;

	(*geometryId) = npstrtoi(&idVal);

	return idVal;
	
}


char* npModelNewTextureId(char* idVal, int* textureId, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPtexmap tex = NULL;

	printf("82934 npModelNewTextureId\n");

	(*textureId) = npstrtoi(&idVal);

	return idVal;
}

void npGeoPrint(pNPgeolist geo, void* dataRef);
void npGeoPrint(pNPgeolist geo, void* dataRef)
{
	pData data = (pData) dataRef;
	
	printf("Geo id %d , extTexId %d, loaded %d ", geo->geometryId, geo->extTexId, geo->loaded );

	if( geo->modelFile != NULL && geo->modelPath != NULL )
		printf("file %s , path %s ,  ", geo->modelFile, geo->modelPath );

	if( geo->modelTextureFile != NULL && geo->modelTexturePath != NULL )
		printf("tex file %s , path %s \n", geo->modelTextureFile, geo->modelTexturePath );

	printf("center (%0f,%0f,%0f)\n", geo->center.x, geo->center.y, geo->center.z);
	printf("rotate (%0f,%0f,%0f)\n", geo->rotate.x, geo->rotate.y, geo->rotate.z);
	printf("scale (%0f,%0f,%0f)\n", geo->scale.x, geo->scale.y, geo->scale.z);

	return;
}

bool npGeoIDisSet2(int* GeoID, void* dataRef);
bool npGeoIDisSet2(int* GeoID, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geo = NULL;
	geo = npGeo(kNPgetGeoIdB, GeoID, NULL, NULL, NULL, dataRef);
	if(geo)
		return true;
	else
		return false;
}

//geo = npGeo(kNPgetGeoId, xGeoId, path, file, dataRef)
bool npGeoIDisSet(int* GeoID, char* path, char* file, void* dataRef);
bool npGeoIDisSet(int* GeoID, char* path, char* file, void* dataRef)
{
	pNPgeolist geo = NULL;
	geo = npGeo(kNPgetGeoId, GeoID, NULL, path, file, dataRef);
	if(geo)
		return true;
	else
		return false;
}

//data->io.gl.geoMap[extGeoId]->loaded != 1
int npGeoMapIDisLoaded(int extGeoId, void* dataRef)
{
	pData data = (pData) dataRef;
	if(data->io.gl.geoMap[extGeoId]->loaded == 1)
		return 1;
	else
		return 0;
}

/*!	Returns internal GeoID given external GeoID
	@param External GeoID
*/
//------------------------------------------------------------------------------
int npGeoMapIDint(int extGeoId, void* dataRef)
{
	pData data = (pData) dataRef;
	return data->io.gl.geoMap[extGeoId]->geometryId;
}

int npGeoMapIDextTexID(int intGeoId, void* dataRef)
{
	pData data = (pData) dataRef;
	return data->io.gl.geoMap[intGeoId]->extTexId;
}

int npGeoMapIDintTexID(int intGeoId, void* dataRef)
{
	pData data = (pData) dataRef;
	return data->io.gl.geoMap[intGeoId]->textureId;
}

pNPgeolist npGeoMapIDis(int extGeoId, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgl gl = &data->io.gl;
	return gl->geoMap[extGeoId];
}

/*
bool npGeoMapIDis(int extGeoId, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgl gl = &data->io.gl;
	bool geoIDis = false;

	if(gl->geoMap[extGeoId])
	{
		geoIDis = true;
	}
	else
	{
		geoIDis = false;
	}

	return geoIDis;
}
*/

pNPtexmap npTexIDset(int* xTexId, char* path, char* file, void* dataRef);
pNPtexmap npTexIDset(int* xTexId, char* path, char* file, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPtexmap texmap = NULL;
	int texCount;

//	texCount == (++data->io.gl.texlistCount);
//	texmap = &data->io.gl.texlist[texCount];

	strcpy(texmap->path, path);
	strcpy(texmap->filename, file);
	sprintf(texmap->filePath, "%s%s", texmap->path, texmap->filename);

	texmap->extTexId = (*xTexId);
//	data->io.gl.texmap[(*xTexId)] = texmap;

	return texmap;
}	

pNPgeolist npGeoIDset(int* xGeoId, char* path, char* file, void* dataRef);
pNPgeolist npGeoIDset(int* xGeoId, char* path, char* file, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeo geo = NULL;
	static int gGeoId = 1000;
	int geoCount;

	geoCount = (++data->io.gl.geoLen);
	geo = &data->io.gl.geolist[geoCount];
			
	strcpy(geo->modelPath, path); strcpy(geo->modelFile, file);
	geo->modelFilePath[0] = '\0';
	sprintf(geo->modelFilePath, "%s%s", geo->modelPath, geo->modelFile);

	if( (*xGeoId) <= 0 )
		geo->geometryId = (*xGeoId) = (++gGeoId);
	else
		geo->geometryId = (++gGeoId);

//	data->io.gl.geoMap[geo->geometryId] = geo;
	data->io.gl.geoMap[(*xGeoId)] = geo;
	
	return geo;
}

pNPgeo npGeoMapGet(int geometryId, void* dataRef);
pNPgeo npGeoMapGet(int geometryId, void* dataRef)
{
	pData data = (pData) dataRef;
	return data->io.gl.geoMap[geometryId];
}

void npGeoMapSetNULL(int geometryId, void* dataRef);
void npGeoMapSetNULL(int geometryId, void* dataRef)
{
	pData data = (pData) data = (pData) dataRef;
	data->io.gl.geoMap[geometryId] = NULL;
}

void npGeoIDunset(int* xGeoId, void* dataRef);
void npGeoIDunset(int* xGeoId, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geo = NULL;

	data->io.gl.geoLen--;
	geo = npGeoMapGet( (*xGeoId), dataRef);
	geo->loaded = 0;
	geo->modelFile[0] = '\0';
	geo->modelFilePath[0] = '\0';
	geo->modelId = 0;
	geo->modelPath[0] = '\0';
	geo->modelTextureFile[0] = '\0';
	geo->modelTextureFilePath[0] = '\0';
	geo->modelTexturePath[0] = '\0';
	geo->textureId = 0;
	npGeoMapSetNULL((*xGeoId), dataRef);
}

void* npGeo(int action, int* xGeoId, int* texId, char* path, char* file, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geoMap = NULL;
	int geoCount = data->io.gl.geoLen;
	int modelID = 0;
	int* intTexId = 0;
	int intId = 0;
	int i = 0;
	//static int gGeoId = 1000;
	char fp[256] = {'\0'};
	pNPgeolist geo = &data->io.gl.geolist[geoCount];
	pNPtexmap tex = NULL;

	geoMap = npGeoMapGet((*xGeoId), dataRef);

	if(!geoMap && (action == kNPgeoLoad))
	{
	}


	if(action == kNPgeoLoad)
	{
		if(geoMap == NULL)
		{
			return geoMap;
		}

		if(geoMap->loaded == 0)
		{
			modelID = npLoadModel(geoMap, dataRef);
			if(modelID)
			{
				npModelStoreDL(((pNPassimp)data->io.assimp)->scene[modelID], geoMap, dataRef);
				geoMap->loaded = 1;
				data->io.gl.numModels++;
			}
			else
			{
				geoMap->loaded = 0;
				npGeo(kNPunsetGeoId, xGeoId, NULL, path, file, dataRef);
				data->io.gl.numModels--;
			}
		}
		return geoMap;
	}

	if(action == kNPgeoLoadTex)
	{
		if(geoMap && geoMap->loaded == 1
			&& geoMap->geometryId > 0)
		{
			sprintf(fp, "%s%s", geoMap->modelTexturePath, geoMap->modelTextureFile);
			geoMap->textureId = npLoadTexture(fp, dataRef);
			return geoMap;
		}

		return geoMap;
	}
	
	if( action == kNPgetGeoIdB)
	{
		return data->io.gl.geoMap[(*xGeoId)];
	}	

#define kNPgeoMapSize 2000
	if( action == kNPgetGeoId)
	{
		for(i = 0; i < kNPgeoMapSize; i++)
		{
			geoMap = data->io.gl.geoMap[i];

			if(geoMap)
			{
				if(i == 1001)
				{
					printf("1001 : %s\n", geoMap->modelPath);
				}
				

				if( geoMap->geometryId == i )
				{
					if(  (strcmp(geoMap->modelPath, path) == 0) &&
						(strcmp(geoMap->modelFile, file) == 0) )
					{
						printf("Already Exists\n");
						(*xGeoId) = i;
						return geoMap;
					}
				}
			}
		}
		return NULL;
		/*
		if(geoMap)
		{
			if(  (strcmp(geoMap->modelPath, path) == 0) &&
				 (strcmp(geoMap->modelFile, file) == 0) )
			{
					printf("Already Exists\n");
					return geoMap;
			}
			else
			{
				printf("Doesn't exist\n");
				(*xGeoId++); 
				return NULL;
			}
		}
		*/
		return NULL;
	}
	
	if( action == kNPunsetGeoId )
	{
		npGeoIDunset(xGeoId, dataRef);
		return NULL;
	}

	if( action == kNPsetGeoId )
	{
		// check if already mapped
	//	if( !(geo = npGeo(kNPgetGeoId, xGeoId, path, file, dataRef)) )
	//	if( !(npGeoIDisSet(xGeoId, path, file, dataRef)) )	
		if( !(npGeoIDisSet2(xGeoId, dataRef)) )	
		{
			// geo not set, set npSetGeo
			geo = npGeoIDset(xGeoId, path, file, dataRef);
			return geo;
		}
		else
		{
			printf("already mapped\n");
			(*xGeoId)++;
			npGeo(kNPsetGeoId, xGeoId, NULL, path, file, dataRef);
		}

	}

	if(action == kNPgeoMapIntTexId)
	{
		//intTexId = npGeoMapIDintTexID((*xGeoId), dataRef);
		(*texId) = npGeoMapIDintTexID((*xGeoId), dataRef);
		return geoMap;
	}

	if(action == kNPgeoMapExtTexId)
	{
		(*texId) = npGeoMapIDextTexID((*xGeoId), dataRef);
		return geoMap;
	}

	if(action == kNPgeoMapIDis)
	{
		geoMap = npGeoMapIDis((*xGeoId), dataRef);
		return geoMap;
	}

	if(action == kNPgeoSetTex)
	{
		//geoMap->modelTexturePath
		strcpy(geoMap->modelTexturePath, path);
		strcpy(geoMap->modelTextureFile, file);
		//data->io.gl.texmap[
		intId = (*texId);
		(*texId) = 0;
		tex = npTex(kNPtexlistAddFile, texId, geoMap->modelTextureFile, geoMap->modelTexturePath, dataRef);
		//tex->intTexId = (*intTexId);
		geo->extTexId = tex->extTexId;
		tex->intTexId = intId;
		tex->loaded = 1;

		return geoMap;
	}

	return NULL;
}

void npLoadExtGeos(void* dataRef)
{
	pData data = (pData) dataRef;
	char fp[256] = {'\0'};
	int i;

	
	for(i = 1000; i < 2000; i++)
	{
		if( data->io.gl.geoMap[i] &&
			data->io.gl.geoMap[i]->loaded == 0
			&&
			data->io.gl.geoMap[i]->modelFile[0] != '\0'
			&&
			data->io.gl.geoMap[i]->modelPath[0] != '\0'
			&&
			data->io.gl.geoMap[i]->textureId == 0)
		{
			npLoadModel(data->io.gl.geoMap[i], dataRef);
		}
	}

}

pNPgeo npModelNew2(char* model_csvline, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeo geo = NULL;
	char* curs = model_csvline;
	char* foundPath = NULL;
	char objectName[256] = {'\0'};
	char fileName[256] = {'\0'};
	char filePath[256] = {'\0'};
	int geometryId = 0;
	int textureId = 0;
	int intId = 0;
	int typeId = 0;
	int i = 0;
	int x = 0;
	NPfloatXYZ center;
	NPfloatXYZ rotate;
	NPfloatXYZ scale;

	curs = npModelNewGeoId(curs, &geometryId, dataRef);
	geo = &data->io.gl.tmpGeoMap[geometryId];
	geo->geometryId = geometryId;
	
	curs = npModelNewTypeId( curs, &typeId, dataRef );

//	curs = npModelNewTextureId(curs, &geo->textureId, dataRef);
	curs = npModelNewTextureId(curs, &geo->extTexId, dataRef);

	curs = npModelNewCenter( curs, &geo->center.x, &geo->center.y, &geo->center.z, dataRef);
	curs = npModelNewCenter( curs, &geo->rotate.x, &geo->rotate.y, &geo->rotate.z, dataRef);
	curs = npModelNewCenter( curs, &geo->scale.x,  &geo->scale.y,  &geo->scale.z,  dataRef);

	curs = npModelNewObjectName(curs, 200, &geo->name[0], dataRef);
	curs =  npModelNewFileName(curs, 50, &geo->modelFile[0], x, dataRef);
//	curs = npModelNewFilePath(curs, 200, &geo->modelFilePath[0], x, dataRef); 
	curs = npModelNewFilePath(curs, 200, &geo->modelPath[0], x, dataRef); 



	return geo;
}


pNPgeo npModelNew(char* model_csvline, void* dataRef)
{	
	pData data = (pData) dataRef;
	pNPgeo geo = NULL;
	char* curs = model_csvline;
	char* foundPath = NULL;
	char objectName[256] = {'\0'};
	char fileName[256] = {'\0'};
	char filePath[256] = {'\0'};
	int geometryId = 0;
	int textureId = 0;
	int intId = 0;
	int typeId = 0;
	int i = 0;
	int x = 0;
	NPfloatXYZ center;
	NPfloatXYZ rotate;
	NPfloatXYZ scale;

	curs = npModelNewGeoId(curs, &geometryId, dataRef);

	curs = npModelNewTypeId( curs, &typeId, dataRef );

	curs = npModelNewTextureId(curs, &textureId, dataRef);

	curs = npModelNewCenter( curs, &center.x, &center.y, &center.z, dataRef);
	curs = npModelNewCenter( curs, &rotate.x, &rotate.y, &rotate.z, dataRef);
	curs = npModelNewCenter( curs, &scale.x,  &scale.y,  &scale.z,  dataRef);

	curs = npModelNewObjectName(curs, 200, &objectName[0], dataRef);
	curs =  npModelNewFileName(curs, 50, &fileName[0], x, dataRef);
	curs = npModelNewFilePath(curs, 200, &filePath[0], x, dataRef); 

//	pNPgeolist npGeo(int action, int xGeoId, char* path, char* file, void* dataRef)
	geo = npGeo(kNPsetGeoId, &geometryId, NULL, filePath, fileName, dataRef);
	if(!geo)
		return NULL;
	else
	{
		geo->center.x = center.x; geo->center.y = center.y; geo->center.z = center.z;
		geo->rotate.x = rotate.x; geo->rotate.y = rotate.y; geo->rotate.z = rotate.z;
		geo->scale.x = scale.x; geo->scale.y = scale.y; geo->scale.z = scale.z;

		

		//if( data->io.gl.extMap[textureId] && (data->io.gl.extMap[textureId]->extTexId > 0) )
		//if( npTexMapIDis(textureId, dataRef) && npTexMapIDext(textureId, dataRef) > 0 )
		if( npTex(kNPtexMapIDis, &textureId, NULL, NULL, dataRef) && npTexMapIDext(textureId, dataRef) > 0 )
		{
			geo->extTexId = textureId;	
		}
		else
		{
			geo->extTexId = textureId;
		}
	}

	if(i == 2000)
	{
		geo = npGetUnusedGeo(dataRef);
	}

	if(filePath[0] == '\0')
	{
		foundPath = npSearchPathsForFile(fileName, dataRef);
		if(foundPath[0] != '\0')
		{
			foundPath[strlen(foundPath)-strlen(fileName)] = '\0';
			strcpy(filePath, foundPath);
		}
	}
/*
	if(geo)
		npGeoPrint(geo, dataRef);
*/	
	return geo;
}



char* npModelNewFilePath(char* stringVal, int maxSize, char* filePath, int i, void* dataRef)
{
	pData data = (pData) dataRef;
	char* curs = stringVal;
	char** read = NULL;
	char* abs = NULL;
	int strLength = 0;
	int len = 0;
	int x = 0;
	strLength = maxSize;

	if (strLength >= 200 )
		strLength = 200;
	
	filePath[0] = '\0';

	npCSVstrncpy(filePath, &curs, strLength);

	i = len = strlen(filePath);
	
	if(filePath[strlen(filePath)-1] != '\\' && len > 0)
	{
		filePath[strlen(filePath)+1] = '\0';
		filePath[strlen(filePath)] = '\\';
	}

	if( npPathIsRel(filePath, dataRef))
	{
		abs = npFilePathRelToAbsNew(filePath, dataRef);
		for(x = 0; x < strlen(abs); x++)
		{
			abs[x] = tolower(abs[x]);
		}

		//abs = npFilePathRelToAbs(filePath, dataRef);
		filePath[0] = '\0';
		strcpy(filePath, abs);
		free(abs);
	}

	len = strlen(filePath);
	for(x = 0; x < len; x++)
		filePath[x] = tolower(filePath[x]);

	return curs;
}




char* npModelNewFileName(char* stringVal, int maxSize, char* fileName, int i, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPgeolist geo = &data->io.gl.geolist[i];
	char* curs = stringVal;
	char** read = NULL;
	int strLength = 0;
	int match = 0;
	char* ext = NULL;
	strLength = maxSize;

	if (strLength >= 200 )
		strLength = 200;
	
	fileName[0] = '\0';

	npCSVstrncpy(fileName, &curs, strLength);

	ext = strstr(fileName, ".");
	if( ext == NULL )
	{
		// model file has no extension, cannot load
		printf("model file has no extension, cannot load\n");
	}
	
	return curs;
}

char* npModelNewObjectName(char* stringVal, int maxSize, char* objectName, void* dataRef)
{
	char* curs = stringVal;
	char** read = NULL;
	int strLength = 0;
	int match = 0;
	strLength = 200;

	if (strLength >= 200 )
		strLength = 200;
	
	objectName[0] = '\0';

	npCSVstrncpy(objectName, &curs, maxSize);

	if( objectName[0] == '\0' )
	{
		
	}

	return curs;
}

//curs = npModelNewTypeId(curs, &typeId, dataRef);
char* npModelNewTypeId(char* csv_typeId, int* typeId, void* dataRef)
{
	(*typeId) = npstrtoi(&csv_typeId);	
	return csv_typeId;
}

void npSetGeoTextures(pNPnode node, void* dataRef)
{
	pData data = (pData) dataRef;
	if(node->geometry > 1000)
	{
		node->textureID = data->io.gl.geoMap[node->geometry]->textureId;
	}
}

void npUpdateGeoListTextures( void* dataRef );
void npUpdateGeoListTextures( void* dataRef )
{
	pData data = (pData) dataRef;
	pNPgeolist geo = NULL;
	pNPtexmap tex = NULL;
	int i = 1001;
	int x = 0;
	int z = 0;

	for(; i <= data->io.gl.geoLen+1000; i++)
	{
		geo = npGeoMapGet(i, dataRef);
		if( geo && (geo->textureId == 0 && geo->extTexId != 0) )
		{
			for(x = 0; x < 2000; x++)
			{
				tex = data->io.gl.texmap[x];
				if(tex){
					if((tex->extTexId == geo->extTexId) && (tex->intTexId != 0) )
					{
						geo->textureId = tex->intTexId;
				//		npTraverseMap(
						//npSetSelectedNodes(kNPtextureID, &geo->textureId, dataRef);
						break;
					}
					else if(tex->loaded == 2) // 2 means that the model file texture path failed, now try same folder as model
					{
						tex->loaded = 3;
					//	sprintf(tex->filePath, "%susr\\global\\images\\%s", data->io.file.appPath, tex->filename);
						sprintf(tex->filePath, "%s%s", geo->modelPath, tex->filename);
						for(z = 0; z < strlen(tex->filePath); z++)
							tex->filePath[z] = tolower(tex->filePath[z]);

						strcpy(tex->path, geo->modelPath);
						for(z = 0; z < strlen(tex->path); z++)
							tex->path[z] = tolower(tex->path[z]);
					}
					else if(tex->loaded == 4) // 4 means that the model file path failed, now try usr glb imgs
					{
						sprintf(tex->filePath, "%susr\\global\\images\\%s", data->io.file.appPath, tex->filename);
						for(z = 0; z < strlen(tex->filePath); z++)
							tex->filePath[z] = tolower(tex->filePath[z]);
						
						sprintf(tex->path, "%susr\\global\\images\\", data->io.file.appPath);
						for(z = 0; z < strlen(tex->path); z++)
							tex->path[z] = tolower(tex->path[z]);
					}
				}
			}

			npTraverseMap(npSetGeoTextures, dataRef);
		}
	}
}

void npUpdateGeoList( void* dataRef );
void npUpdateGeoList( void* dataRef )
{
	pData data = (pData) dataRef;
	pNPgeolist geo = NULL;
	int i = 1001;


	npUpdateGeoListTextures( dataRef );

//	if( data->io.file.loading == false)
//	{
	for(; i <= (1000 + data->io.gl.geoLen); i++)
	{
		geo = npGeoMapGet(i, dataRef);
		if(geo && geo->modelFilePath[0] != '\0' && geo->loaded == 0)
		{
			printf("To Load : %p | i : %d | geometryId : %d ", geo, i, geo->geometryId );
			printf("geo->modelFile : %s\n", geo->modelFile);
			//npGeo(kNPgeoLoad, &xGeoId, path, fileName, dataRef);
			printf("npUpdateGeoList , i : %d\n", i);
			geo = npGeo(kNPgeoLoad, &geo->geometryId, NULL, geo->modelPath, geo->modelFile, dataRef);
			//printf("after npGeoId\n");
			if(geo)
				geo->loaded = 1;
		}
	}
//	}

	return;
}

/** Imports an assimp scene from a file
	@param filePath, full path to file
	@param dataRef is a global map reference instance.
*/
const struct aiScene* npModelImport(char* filePath, void* dataRef)
{
	const struct aiScene* scene = NULL;

	scene = aiImportFile((const char*)filePath, 0);

	if( scene == NULL )
		printf("\naiImportFile returned null");

	return scene;
}
