/* -----------------------------------------------------------------------------
*
*  npusgs.c
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2016 by Shane Saxon - saxon@openantz.com
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named LICENSE.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#include "npusgs.h"



//pNPusgs usgs;

void npInitUSGS(void* dataRef)
{
	pData data = (pData) dataRef;
	pNPusgs usgs = NULL;
	
	data->io.usgs = (pNPusgs)malloc(sizeof(NPusgs));
	usgs = ((pNPusgs)data->io.usgs);
	if(usgs)
	{
		usgs->eqLink[0] = '\0';
	//	strcpy(usgs->eqLink, "http://earthquake.usgs.gov/earthquakes/feed/v1.0/summary/4.5_day.quakeml");
	//	strcpy(usgs->eqLink, "http://earthquake.usgs.gov/earthquakes/feed/v1.0/summary/1.0_hour.atom");
		strcpy(usgs->eqLink, "http://earthquake.usgs.gov/earthquakes/feed/v1.0/summary/1.0_day.atom");
		usgs->action = 1;
	}

	
}

size_t npUSGSWriteMemoryCallback(void* contents, size_t size, size_t nmemb, void* userp)
{
	size_t realSize = size * nmemb;
	pData data = (pData) userp;
	pNPatom atom = &data->io.file.atom;
	pNPatomRequestResponse rR = &atom->rR;
//	pNPatomRequestResponse rR = &data->io.file.ato
//	pNPgithub github = &data->io.github;
//	pNPgithubRequestResponse rR = &github->rR[github->requestResponse_index];	

	
//	printf("npUSGSWriteMemoryCallback : %d\n", rR->responseSize);

	rR->response = realloc(rR->response, rR->responseSize + realSize + 1);
	if(rR->response == NULL)
	{
		printf( "rR->response == NULL\n" );
		return 0; /// out of memory
	}

//	printf("contents : %s\n%d\n", (char*)contents, github->requestResponse_index);
	memcpy(&rR->response[rR->responseSize], contents, realSize);
	rR->responseSize += realSize;

	/// null terminate it
	rR->response[rR->responseSize] = '\0';

	return realSize;
}

#define kNPusgsRequest 1
void npUpdateUSGS(void* dataRef)
{
	pData data = (pData) dataRef;
	pNPusgs usgs = (pNPusgs)data->io.usgs;
	/*	action :
		- kNPusgsGet
		- kNPusgsFeed

	*/
	usgs->action = kNPusgsRequest;
	if(usgs->action == kNPusgsRequest)
	{
		// set curl response callback function
		//err = npCURLgetUrl(curl, github->rR[request_index].requestUrl, 0, dataRef);
		npCURL_easySetOptWriteFunction(&data->io.curl, npUSGSWriteMemoryCallback, dataRef);
		npCURLgetUrl(&data->io.curl, usgs->eqLink, 0, dataRef);
	}

}

void npTestUSGS(void* dataRef)
{

}

void npCloseUSGS(void* dataRef)
{

}